import { Parser } from "binary-parser"
import { IMessageHeader, IMessageQuestion, MessageRecourceRecord, ErrorCodes } from "./types"
import { Serializer, Serializeable } from "./serializeable";
import { Question } from "./question";
import { RecourceRecord } from "./record";
import { Header } from "./header";

export const headerParser: Parser = new Parser()
   .endianess("big")
   .uint16("ID")
   .bit1("QR")
   .bit4("OPCODE")
   .bit1("AA")
   .bit1("TC")
   .bit1("RD")
   .bit1("RA")
   .bit1("Z")
   .bit1("AD")
   .bit1("CD")
   .bit4("RCODE")
   .uint16("QDCOUNT")
   .uint16("ANCOUNT")
   .uint16("NSCOUNT")
   .uint16("ARCOUNT")

export const labelParser = new Parser()
   .endianess("big")
   .uint8("dataLength")
   .string("name", {
      length: "dataLength",
      encoding: "ascii"
   })

export const questionParser = new Parser()
   .endianess("big")
   .array("QNAME", {
      type: labelParser,
      readUntil: (item: any, buffer) => {
         if (item.dataLength <= 0)
            return true;
      },
      formatter: (value: { dataLength: number, name: string }[]) => {
         return value.map(e => e.name).join(".").slice(0, -1);
      }
   })
   .uint16("QTYPE")
   .uint16("QCLASS")


const packetParser = new Parser()
   .endianess("big")
   .nest("header", { type: headerParser })
   .array("questions", {
      type: questionParser,
      length: "header.QDCOUNT"
   })

export function parseInput(data: Buffer): { header: IMessageHeader, questions: IMessageQuestion[] } {
   let res = packetParser.parse(data);
   return <any>res;
}

export class Request {
   private _header: Header;
   get header() {
      return { ...this._header };
   }

   private _questions: Question[];
   get questions() {
      return this._questions.map(e => new Question(e));
   }

   private _answers: Serializer[] = [];
   addAnswer(rr: RecourceRecord) {
      this._answers.push(rr.lock())
   }

   private _authorities: Serializer[] = [];
   addAuthorities(rr: RecourceRecord) {
      this._authorities.push(rr.lock())
   }

   private _additionals: Serializer[] = [];
   addAdditionals(rr: RecourceRecord) {
      this._additionals.push(rr.lock())
   }

   constructor(packet: Buffer, public sender: string, private max_size = 0) {
      let parsed = parseInput(packet);

      this._header = new Header(parsed.header);
      this._header.AD = 0;
      this._header.RCODE = ErrorCodes.NoError;
      this._header.RA = this._header.RD;
      this._questions = parsed.questions.map(e => new Question(e));
   }

   error(error: ErrorCodes) {
      if (this._header.RCODE === ErrorCodes.NoError)
         this._header.RCODE = error;
   }

   noRecursion() {
      this._header.RA = 0;
   }

   serialize() {
      this._header.AA = 1;
      this._header.ANCOUNT = this._answers.length;
      this._header.ARCOUNT = this._additionals.length;
      this._header.NSCOUNT = this._authorities.length;
      this._header.QR = 1;

      let header = this._header.lock();
      let questions = this._questions.map(e => e.lock())
      let answers = this._answers;
      let authority = this._authorities;
      let additional = this._additionals;

      let length = header.length;
      questions.forEach(e => length += e.length);
      answers.forEach(e => length += e.length)
      authority.forEach(e => length += e.length)
      additional.forEach(e => length += e.length)

      if (this.max_size && length > this.max_size) {
         this._header.TC = 1;

         //Will ignore data, that exceeds length
         length = this.max_size;
      }


      let data = Buffer.alloc(length)
      let offset = 0;

      let append = (ser: Serializer) => {
         if (offset <= length) {
            offset = ser.serialize(data, offset);
         }
      }
      append(header)
      questions.forEach(append)
      answers.forEach(append)
      authority.forEach(append)
      additional.forEach(append)
      return data;
   }
}
